/**
 * Copyright (C) 2013
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 */
package org.n52.sos.binding.rest;

import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.xmlbeans.XmlException;
import org.n52.sos.binding.Binding;
import org.n52.sos.binding.rest.decode.RestDecoder;
import org.n52.sos.binding.rest.encode.RestEncoder;
import org.n52.sos.binding.rest.requests.BadRequestException;
import org.n52.sos.binding.rest.requests.RequestHandler;
import org.n52.sos.binding.rest.requests.RestRequest;
import org.n52.sos.binding.rest.requests.RestResponse;
import org.n52.sos.binding.rest.resources.OptionsRequestHandler;
import org.n52.sos.binding.rest.resources.OptionsRestRequest;
import org.n52.sos.binding.rest.resources.ServiceEndpointRequest;
import org.n52.sos.binding.rest.resources.ServiceEndpointRequestHandler;
import org.n52.sos.binding.rest.resources.capabilities.CapabilitiesRequestHandler;
import org.n52.sos.binding.rest.resources.capabilities.ICapabilitiesRequest;
import org.n52.sos.binding.rest.resources.features.FeaturesRequest;
import org.n52.sos.binding.rest.resources.features.FeaturesRequestHandler;
import org.n52.sos.binding.rest.resources.observations.IObservationsRequest;
import org.n52.sos.binding.rest.resources.observations.ObservationsRequestHandler;
import org.n52.sos.binding.rest.resources.offerings.OfferingsRequest;
import org.n52.sos.binding.rest.resources.offerings.OfferingsRequestHandler;
import org.n52.sos.binding.rest.resources.sensors.ISensorsRequest;
import org.n52.sos.binding.rest.resources.sensors.SensorsRequestHandler;
import org.n52.sos.decode.Decoder;
import org.n52.sos.encode.Encoder;
import org.n52.sos.encode.EncoderKey;
import org.n52.sos.encode.XmlEncoderKey;
import org.n52.sos.exception.CodedException;
import org.n52.sos.exception.OwsExceptionReportEncodingFailedException;
import org.n52.sos.exception.ows.MissingParameterValueException;
import org.n52.sos.exception.ows.NoApplicableCodeException;
import org.n52.sos.exception.ows.OwsExceptionCode;
import org.n52.sos.ogc.ows.OwsExceptionReport;
import org.n52.sos.response.ServiceResponse;
import org.n52.sos.service.Configurator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * TODO catch all exception and return useful HTTP codes
 * 
 * @author <a href="mailto:e.h.juerrens@52north.org">Eike Hinderk
 *         J&uuml;rrens</a>
 */
public class RestBinding extends Binding {

    private static final Logger LOGGER = LoggerFactory.getLogger(RestBinding.class);

    private final Set<String> conformanceClasses;

    private final Constants bindingConstants;

    public RestBinding() {
        bindingConstants = Constants.getInstance();
        conformanceClasses = new HashSet<String>(0);
        conformanceClasses.add(bindingConstants.getConformanceClass());
    }

    @Override
    public Set<String> getConformanceClasses()
    {
        return conformanceClasses;
    }

    @Override
    public String getUrlPattern()
    {
        return bindingConstants.getUrlPattern();
    }

    @Override
    public ServiceResponse doOptionsOperation(final HttpServletRequest request) throws OwsExceptionReportEncodingFailedException
    {
        if (isPathInfoAvailable(request)) {
            return doGetOperation(request);
        } else {
            return super.doOptionsOperation(request);
        }
    }

    private boolean isPathInfoAvailable(final HttpServletRequest request)
    {
        return (request != null) && (request.getPathInfo() != null) && !request.getPathInfo().isEmpty();
    }

    @Override
    public ServiceResponse doDeleteOperation(final HttpServletRequest request) throws OwsExceptionReportEncodingFailedException
    {
        return doGetOperation(request);
    }

    @Override
    public ServiceResponse doPutOperation(final HttpServletRequest request) throws OwsExceptionReportEncodingFailedException
    {
        return doGetOperation(request);
    }

    @Override
    public ServiceResponse doPostOperation(final HttpServletRequest request) throws OwsExceptionReportEncodingFailedException
    {
        return doGetOperation(request);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.n52.sos.binding.rest.Binding#doGetOperation(javax.servlet.http.
     * HttpServletRequest)
     * 
     * INPUT: /rest/RESOURCE{/id|?kvps}
     * 
     * EXAMPLE: /rest/sensors/0815 => OGC::DescribeSensor in format SensorML
     * 1.0.1 /rest/features?bbox=52.0,7.0,50.0,8.0 for all features in the given
     * BBox
     */
    @Override
    public ServiceResponse doGetOperation(final HttpServletRequest request) throws OwsExceptionReportEncodingFailedException
    {
        try {
            LOGGER.debug(String.format("Start handling of REST request. URI:%s",request.getRequestURI()));
            ServiceResponse response = null;

            // Decode the request
            final RestRequest restRequest = decodeHttpRequest(request); 
            LOGGER.debug("Rest request decoded to {}", 
                    restRequest!=null?restRequest.getClass().getName():null);

            // Handle the request
            final RestResponse restResponse = handleRestRequest(restRequest);
            LOGGER.debug("Rest request handled. DeleteObservationResponse received: {}",
                    restResponse!=null?restResponse.getClass().getName():null);

            // Encode the response
            response = encodeRestResponse(restResponse);
            LOGGER.debug("Rest response encoded. DeleteObservationResponse received: {}",
                    response!=null?response.getClass().getName():null);
            
            LOGGER.debug("Handling of REST request finished. Returning response to web tier");

            return response;
        } catch (final OwsExceptionReport owsER) {
            LOGGER.error("Error while processing rest request. Exception thrown: {}",
                    owsER.getClass().getSimpleName());
            return encodeIntoServiceResponse(owsER);
        }
    }
    
    private ServiceResponse encodeIntoServiceResponse(final OwsExceptionReport owsER) throws OwsExceptionReportEncodingFailedException
    {
        if ((owsER.getExceptions() != null) && (owsER.getExceptions().size() > 0))
        {
            final ServiceResponse exceptionResponse = encodeOwsExceptionReport(owsER,true);
            
            for (final CodedException owsE : owsER.getExceptions())
            {
                int httpResponseCode = HttpServletResponse.SC_INTERNAL_SERVER_ERROR;
                if (owsE.getCode().equals(OwsExceptionCode.OperationNotSupported))
                {
                    if (isOperationNotSupportedException(owsE))
                    {
                        httpResponseCode = HttpServletResponse.SC_BAD_REQUEST;
                    }
                }
                else if (owsE.getCode().equals(OwsExceptionCode.NoApplicableCode))
                {
                    if (isMethodeNotAllowedExceptionForResourceType(owsE))
                    {
                        httpResponseCode = HttpServletResponse.SC_METHOD_NOT_ALLOWED;
                    }
                    else if (owsE.getMessage().indexOf(bindingConstants.getErrorMessageWrongContentType()) != -1)
                    {
                        httpResponseCode = HttpServletResponse.SC_UNSUPPORTED_MEDIA_TYPE;
                    }
                    else if (owsE.getMessage().indexOf(bindingConstants.getErrorMessageWrongContentTypeInAcceptHeader()) != -1)
                    {
                        httpResponseCode = HttpServletResponse.SC_NOT_ACCEPTABLE;
                    }
                    // TODO if this working?
                    else if (owsE.getCause() instanceof BadRequestException)
                    {
                        httpResponseCode = HttpServletResponse.SC_BAD_REQUEST;
                    }
                    else if ((owsE.getMessage().indexOf("HTTP method") > -1) && (owsE.getMessage().indexOf("not allowed") > -1))
                    {
                    	httpResponseCode = HttpServletResponse.SC_METHOD_NOT_ALLOWED;
                    }
                }
                else if (owsE.getCode().equals(OwsExceptionCode.InvalidParameterValue))
                {
                    httpResponseCode = HttpServletResponse.SC_BAD_REQUEST;
                }
                exceptionResponse.setHttpResponseCode(httpResponseCode);
                return exceptionResponse;
            }
        }
        return encodeOwsExceptionReport(owsER, true);
    }

    private boolean isOperationNotSupportedException(final CodedException owsE)
    {
        return (owsE.getMessage() != null) && !owsE.getMessage().isEmpty() && 
                (owsE.getMessage().indexOf(bindingConstants.getSosErrorMessageOperationNotSupportedStart()) != -1) &&
                        (owsE.getMessage().indexOf(bindingConstants.getSosErrorMessageOperationNotSupportedEnd()) != -1);
    }

    private boolean isMethodeNotAllowedExceptionForResourceType(final CodedException owsE)
    {
        return (owsE.getMessage() != null) && !owsE.getMessage().isEmpty() &&
                (owsE.getMessage().indexOf(bindingConstants.getHttpOperationNotAllowedForResourceTypeMessagePart()) != -1);
    }
    
    private RestRequest decodeHttpRequest(final HttpServletRequest request) throws OwsExceptionReport
    {
        final Decoder<RestRequest, HttpServletRequest> decoder = getDecoder();

        if (decoder == null) {
            final String exceptionText = String.format("No decoder implementation is available for RestBinding and namespace \"%s\"!",
                    bindingConstants.getEncodingNamespace());
            LOGGER.debug(exceptionText);
            throw new NoApplicableCodeException().withMessage(exceptionText);
        }

        // 2 decode request
        final RestRequest restRequest = decoder.decode(request);

        if (restRequest == null) {
            final String exceptionText = "Decoding of request failed but no exception is thrown.";
            LOGGER.debug(exceptionText);
            throw new NoApplicableCodeException().withMessage(exceptionText);
        }
        return restRequest;
    }

    private ServiceResponse encodeRestResponse(final RestResponse restResponse) throws OwsExceptionReport
    {
        ServiceResponse response;
        final RestEncoder encoder = getEncoder();

        if (encoder == null) {
            final String exceptionText = String.format("No encoder implementation is available for RestBinding and namespace \"%s\"!",
                    bindingConstants.getEncodingNamespace());
            LOGGER.debug(exceptionText);
            throw new NoApplicableCodeException().withMessage(exceptionText);
        }
        
        response = encoder.encode(restResponse);

        if (response == null) {
            final String exceptionText = String.format("Encoding of response \"%s\" failed with encoder \"%s\" but no exception is thrown.",
                    restResponse.getClass().getName(),
                    encoder.getClass().getName());
            LOGGER.debug(exceptionText);
            throw new NoApplicableCodeException().withMessage(exceptionText);
        }
        return response;
    }

    private RestResponse handleRestRequest(final RestRequest restRequest) throws OwsExceptionReport
    {
        // 3 get request handler
        final RequestHandler requestHandler = getRequestHandler(restRequest);
        LOGGER.debug("RequestHandler of type {} found for RestRequest of type {}",
                requestHandler!=null?requestHandler.getClass().getName():requestHandler,
                        restRequest!=null?restRequest.getClass().getName():restRequest);
        
        RestResponse sdcResponse = null;
        try {
            sdcResponse = requestHandler.handleRequest(restRequest);
            if (sdcResponse == null) {
                final String exceptionText = "Processing of rest request failed but no exception is thrown.";
                LOGGER.debug(exceptionText);
                throw new NoApplicableCodeException().withMessage(exceptionText);
             }
        }catch(final XmlException xe) {
            final String exceptionText = String.format("Processing of rest request response failed. Exception thrown: %s",
                    xe.getMessage());
            LOGGER.debug(exceptionText,xe);
            throw new NoApplicableCodeException().withMessage(exceptionText).causedBy(xe);
        } catch (final IOException e) {
            final String exceptionText = String.format("Processing of rest request response failed. Exception thrown: %s",
                    e.getMessage());
            LOGGER.debug(exceptionText,e);
            throw new NoApplicableCodeException().withMessage(exceptionText).causedBy(e);
        }
        return sdcResponse;
    }

    @SuppressWarnings("rawtypes")
    private RestEncoder getEncoder() throws OwsExceptionReport
    {
    	final EncoderKey key = new XmlEncoderKey(bindingConstants.getEncodingNamespace(), RestResponse.class);
        final Encoder encoder = getConfigurator().getCodingRepository().getEncoder(key);
        if (encoder instanceof RestEncoder) {
            return (RestEncoder) encoder;
        }
        return null;
    }

    private RequestHandler getRequestHandler(final RestRequest restRequest)
            throws OwsExceptionReport
    {
        if ((restRequest != null)) {
            if (restRequest instanceof ISensorsRequest)
            {
                return new SensorsRequestHandler();
            }
            else if (restRequest instanceof IObservationsRequest) 
            {
                return new ObservationsRequestHandler();
            }
            else if (restRequest instanceof ICapabilitiesRequest)
            {
                return new CapabilitiesRequestHandler();
            } 
            else if (restRequest instanceof OfferingsRequest)
            {
                return new OfferingsRequestHandler();
            } 
            else if (restRequest instanceof FeaturesRequest)
            {
                return new FeaturesRequestHandler();
            }
            else if (restRequest instanceof OptionsRestRequest)
            {
                return new OptionsRequestHandler();
            }
            else if (restRequest instanceof ServiceEndpointRequest)
            {
                return new ServiceEndpointRequestHandler();
            }
        }
        throw new MissingParameterValueException(bindingConstants.getResourceType());
    }

    @SuppressWarnings("unchecked")
    private Decoder<RestRequest, HttpServletRequest> getDecoder() throws OwsExceptionReport {
        final Set<Decoder<?, ?>> decoders = getConfigurator().getCodingRepository().getDecoders();
        for (@SuppressWarnings("rawtypes") final Decoder decoder : decoders) {
            if (decoder instanceof RestDecoder) {
                return decoder;
            }
        }
        return null;
    }

	private Configurator getConfigurator()
	{
		return Configurator.getInstance();
	}

}
