/**
 * Copyright (C) 2013
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 */
package org.n52.sos.binding.rest.encode;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletResponse;

import net.opengis.sosREST.x10.LinkType;
import net.opengis.sosREST.x10.ResourceCollectionType;

import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlOptions;
import org.n52.sos.binding.rest.requests.RestResponse;
import org.n52.sos.exception.ows.NoApplicableCodeException;
import org.n52.sos.exception.ows.concrete.EncoderResponseUnsupportedException;
import org.n52.sos.ogc.OGCConstants;
import org.n52.sos.ogc.om.OMConstants;
import org.n52.sos.ogc.om.features.SFConstants;
import org.n52.sos.ogc.ows.OwsExceptionReport;
import org.n52.sos.ogc.sos.Sos1Constants;
import org.n52.sos.ogc.sos.SosConstants;
import org.n52.sos.response.ServiceResponse;
import org.n52.sos.util.W3CConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author <a href="mailto:e.h.juerrens@52north.org">Eike Hinderk J&uuml;rrens</a>
 *
 */
public abstract class ResourceEncoder extends RestEncoder {
    
    private static final Logger LOGGER = LoggerFactory.getLogger(ResourceEncoder.class);

    public  abstract ServiceResponse encodeRestResponse(RestResponse objectToEncode) throws OwsExceptionReport;
    
    protected String createHrefForResourceType(String resourceType)
    {
        return bindingConstants.getServiceUrl()
                .concat(bindingConstants.getUrlPattern())
                .concat("/")
                .concat(resourceType);
    }
    
    protected String createHrefForResourceAndIdentifier(String resourceType,
            String identifier)
    {
        return createHrefForResourceType(resourceType)
                .concat("/")
                .concat(identifier);
    }
    
    protected String createHrefForResourceTypeAndQueryString(String resourceType, String queryString) {
        return createHrefForResourceType(resourceType)
                .concat("?")
                .concat(queryString);
    }
    
    protected void setValuesOfLinkToDynamicResource(LinkType xb_RestLink, String resourceQueryString, String relation, String resourceTypeIdentifier)
    {
        setLinkValues(xb_RestLink,
                createRelationWithNamespace(relation),
                createHrefForResourceTypeAndQueryString(resourceTypeIdentifier, resourceQueryString),
                bindingConstants.getContentTypeDefault());
    }
    
    protected void setValuesOfLinkToGlobalResource(LinkType xb_Link,
            String relationIdentifier,
            String resourceType)
    {
        setLinkValues(xb_Link,
                createRelationWithNamespace(relationIdentifier),
                createHrefForResourceType(resourceType),
                bindingConstants.getContentTypeDefault());
    }

    protected void setValuesOfLinkToUniqueResource(LinkType xb_RestLink, String resourceId, String relation, String resourceType)
    {
        setLinkValues(xb_RestLink,
                createRelationWithNamespace(relation),
                createHrefForResourceAndIdentifier(resourceType, resourceId),
                bindingConstants.getContentTypeDefault());
    }
    
    protected void setLinkValues(LinkType xb_Link, String rel, String href, String type) {
        xb_Link.setRel(rel);
        xb_Link.setHref(href);
        xb_Link.setType(type);
    }
    
    protected String createRelationWithNamespace(String relation)
    {
        return bindingConstants.getEncodingNamespace().concat("/").concat(relation);
    }
    
    protected void setOfferingLinks(XmlObject xb_AnyType,
            List<String> offeringIdentifiers)
    {
        if (xb_AnyType instanceof net.opengis.sosREST.x10.CapabilitiesType || xb_AnyType instanceof ResourceCollectionType) {
            LinkType xb_RestLink = null;

            for (String offeringIdentifier : offeringIdentifiers) {
                if (xb_AnyType instanceof net.opengis.sosREST.x10.CapabilitiesType) {
                    xb_RestLink = ((net.opengis.sosREST.x10.CapabilitiesType)xb_AnyType).addNewLink();
                } else {
                    xb_RestLink = ((ResourceCollectionType)xb_AnyType).addNewLink();
                }
                setValuesOfLinkToUniqueResource(xb_RestLink,
                        offeringIdentifier,
                        bindingConstants.getResourceRelationOfferingGet(),
                        bindingConstants.getResourceOfferings());
            }
            
        }
    }
    
    protected ServiceResponse createServiceResponseFromXBDocument(XmlObject xb_RestDoc,
            String resourceType,
            int httpResponseCode,
            boolean isResourceCollection,
            boolean isGlobalResource) throws OwsExceptionReport
    {
        return createServiceResponseFromXBDocument(xb_RestDoc,
                resourceType,
                null,
                httpResponseCode,
                isResourceCollection,
                isGlobalResource);
    }
    
    protected ServiceResponse createServiceResponseFromXBDocument(XmlObject xb_RestDoc,
            String resourceType,
            Map<String,String> additionalHeaders,
            int httpResponseCode,
            boolean isResourceCollection,
            boolean isGlobalResource) throws OwsExceptionReport
    {
        ByteArrayOutputStream bAOS = new ByteArrayOutputStream();
        try {
            xb_RestDoc.save(bAOS, getDefaultXMLOptions());
            int status = httpResponseCode;
            if (httpResponseCode < 100 || httpResponseCode > 505) {
                status = HttpServletResponse.SC_OK;
            }
            
            ServiceResponse response = new ServiceResponse(bAOS,
                    bindingConstants.getContentTypeDefault(),
                    false,
                    true,
                    status);
            setAllowHeadersForResourceType(response,resourceType,isResourceCollection,isGlobalResource);
            
            if (additionalHeaders != null && additionalHeaders.size() > 0) {
                for (String headerIdentifier : additionalHeaders.keySet()) {
                    if (headerIdentifier != null && !headerIdentifier.isEmpty()) {
                        String value = additionalHeaders.get(headerIdentifier);
                        if (value != null) {
                            response.addHeaderKvpEntry(headerIdentifier, value);
                        }
                    }
                }
            }

            return response;
        } catch (IOException ioe) {
            String exceptionText = String.format("Encoding of result failed. Exception: %s",
                    ioe.getMessage());
            LOGGER.debug(exceptionText);
            throw new NoApplicableCodeException().
            	causedBy(ioe).
            	withMessage(exceptionText);
        }
    }
    
    private void setAllowHeadersForResourceType(ServiceResponse response,
            String resourceType,
            boolean isResourceCollection,
            boolean isGlobalResource)
    {
        response.addHeaderKvpEntry(
                bindingConstants.getHttpHeaderIdentifierAccessControlAllowMethods(),
                getAllowedHttpMethodsForResourceType(resourceType,isResourceCollection,isGlobalResource));
        
        response.addHeaderKvpEntry(
                bindingConstants.getHttpHeaderIdentifierAllow(),
                getAllowedHttpMethodsForResourceType(resourceType,isResourceCollection,isGlobalResource));
    }
    
    private String getAllowedHttpMethodsForResourceType(String bindingOperation,
            boolean isResourceCollection,
            boolean isGlobalResource)
    {
        StringBuffer allowedOperations = new StringBuffer("");
        allowHttpGet(allowedOperations);
        appendAllowHttpOptions(allowedOperations);
        
        if (bindingOperation.equalsIgnoreCase(bindingConstants.getResourceObservations()))
        {
            if (isGlobalResource && !isResourceCollection)
            {
                appendAllowHttpPost(allowedOperations);
            }
            else if (!isGlobalResource && !isResourceCollection)
            {
                appendAllowHttpDelete(allowedOperations);
            }
        }
        else if (bindingOperation.equalsIgnoreCase(bindingConstants.getResourceSensors()))
        {
            if (!isGlobalResource && !isResourceCollection)
            {
                appendAllowHttpPut(allowedOperations);
                appendAllowHttpDelete(allowedOperations);
            }
            else if (isGlobalResource && isResourceCollection)
            {
                appendAllowHttpPost(allowedOperations);
            }
        }
        return allowedOperations.toString();
    }
    
    private void appendAllowHttpDelete(StringBuffer allowedOperations)
    {
        allowedOperations.append(", ").append(bindingConstants.getHttpDeleteMethodName());
    }
    
    private void appendAllowHttpOptions(StringBuffer allowedOperations)
    {
        allowedOperations.append(", ").append(bindingConstants.getHttpOptionsMethodName());
    }

    private void appendAllowHttpPost(StringBuffer allowedOperations)
    {
        allowedOperations.append(", ").append(bindingConstants.getHttpPostMethodName());
    }
    
    private void appendAllowHttpPut(StringBuffer allowedOperations)
    {
        allowedOperations.append(", ").append(bindingConstants.getHttpPutMethodName());
    }

    private void allowHttpGet(StringBuffer allowedOperations)
    {
        allowedOperations.append(bindingConstants.getHttpGetMethodName());
    }
    
    protected XmlOptions getDefaultXMLOptions()
    {
        XmlOptions bindingXmlOptions = new XmlOptions();
        bindingXmlOptions.setSaveAggressiveNamespaces();
        bindingXmlOptions.setSavePrettyPrint();
        bindingXmlOptions.setSaveNamespacesFirst();
        
        // TODO change, if method is available from SOS::API
        Map<String,String> sosCorePrefixMap = /*XmlOptionsHelper.getInstance().*/getPrefixMap();
        sosCorePrefixMap.put(bindingConstants.getEncodingNamespace(), bindingConstants.getEncodingPrefix());
        bindingXmlOptions.setSaveSuggestedPrefixes(sosCorePrefixMap);
        bindingXmlOptions.setSaveImplicitNamespaces(sosCorePrefixMap);
        
        return bindingXmlOptions;
                
    }
    
    // TODO remove, if method is available from SOS::API
    private Map<String,String> getPrefixMap(){
        Map<String, String> lPrefixMap = new Hashtable<String, String>();
      lPrefixMap.put(OGCConstants.NS_OGC, OGCConstants.NS_OGC_PREFIX);
      lPrefixMap.put(OMConstants.NS_OM, OMConstants.NS_OM_PREFIX);
      lPrefixMap.put(SFConstants.NS_SA, SFConstants.NS_SA_PREFIX);
      lPrefixMap.put(Sos1Constants.NS_SOS, SosConstants.NS_SOS_PREFIX);
      lPrefixMap.put(W3CConstants.NS_XLINK, W3CConstants.NS_XLINK_PREFIX);
      lPrefixMap.put(W3CConstants.NS_XSI, W3CConstants.NS_XSI_PREFIX);
      lPrefixMap.put(W3CConstants.NS_XS, W3CConstants.NS_XS_PREFIX);
      return lPrefixMap;
    }

    protected OwsExceptionReport createResponseNotSupportedException(String expectedClassString, Object receivedObject)
    {
    	return new EncoderResponseUnsupportedException().
    			withMessage("Received RestResponse is not a '%s' but of type '%s'", 
                        expectedClassString,
                        receivedObject!=null?receivedObject.getClass().getName():"null");
    }

    protected void addLocationHeader(ServiceResponse response,
            String resourceType)
    {
        addLocationHeader(response, null, resourceType);
    }

    protected void addLocationHeader(ServiceResponse response,
            String resourceId,
            String resourceType)
    {
        String location = null;
        if (resourceId == null)
        {
            location = createHrefForResourceType(resourceType);
        } 
        else
        {
            location = createHrefForResourceAndIdentifier(resourceType,
                    resourceId);
        }
        response.addHeaderKvpEntry(
                bindingConstants.getHttpHeaderIdentifierLocation(),
                location);
    }

    protected ServiceResponse createNoContentResponse(String resourceType,
            boolean isResourceCollection,
            boolean isGlobalResource)
    {
        return createContentlessResponse(resourceType, HttpServletResponse.SC_NO_CONTENT,isResourceCollection,isGlobalResource);
    }
    
    protected ServiceResponse createContentlessResponse(String resourceType,
            int httpStatusCode,
            boolean isResourceCollection,
            boolean isGlobalResource)
    {
        ServiceResponse response = new ServiceResponse(bindingConstants.getContentTypeDefault(),httpStatusCode);
        setAllowHeadersForResourceType(response, resourceType, isResourceCollection, isGlobalResource);
        return response; 
    }

}
