/**
 * Copyright (C) 2013
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 */
package org.n52.sos.binding.rest.resources.sensors;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import net.opengis.sensorML.x101.SensorMLDocument;
import net.opengis.sensorML.x101.SystemType;
import net.opengis.swes.x20.AbstractContentsType.Offering;
import net.opengis.swes.x20.DescribeSensorResponseDocument;
import net.opengis.swes.x20.DescribeSensorResponseType.Description;
import net.opengis.swes.x20.SensorDescriptionType;

import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;
import org.n52.sos.binding.rest.requests.ResourceNotFoundResponse;
import org.n52.sos.binding.rest.requests.RestRequest;
import org.n52.sos.binding.rest.requests.RestResponse;
import org.n52.sos.exception.CodedException;
import org.n52.sos.exception.ows.NoApplicableCodeException;
import org.n52.sos.exception.ows.OwsExceptionCode;
import org.n52.sos.ogc.ows.OwsExceptionReport;
import org.n52.sos.ogc.sensorML.SensorMLConstants;
import org.n52.sos.ogc.sos.SosConstants;
import org.n52.sos.request.DescribeSensorRequest;
import org.n52.sos.response.ServiceResponse;
import org.n52.sos.service.operator.ServiceOperator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author <a href="mailto:e.h.juerrens@52north.org">Eike Hinderk J&uuml;rrens</a>
 */
public class SensorsGetRequestHandler extends SensorsRequestHandler {

    private static final Logger LOGGER = LoggerFactory.getLogger(SensorsGetRequestHandler.class);

    @Override
    public RestResponse handleRequest(RestRequest sensorsHttpGetRequest) throws OwsExceptionReport, XmlException, IOException
    {
        if (sensorsHttpGetRequest != null) {
            if (sensorsHttpGetRequest instanceof GetSensorByIdRequest)
            {
                // Case A: with ID
                return handleGetSensorByIdRequest((GetSensorByIdRequest)sensorsHttpGetRequest);
            }
            else if (sensorsHttpGetRequest instanceof GetSensorsRequest)
            {
                // Case C: global resource
                return handleGetSensorsRequest((GetSensorsRequest)sensorsHttpGetRequest);
            }
        }
        throw logRequestTypeNotSupportedByThisHandlerAndCreateException(sensorsHttpGetRequest,this.getClass().getName());
    }

    private SensorsGetResponse handleGetSensorsRequest(GetSensorsRequest getSensorsRequest) throws OwsExceptionReport, XmlException, IOException
    {
        Offering[] xb_offerings = getOfferingsFromSosCore(getSensorsRequest.getCapabilitiesRequest());
        Set<String> sensorIds = new HashSet<String>();
        String[] sensorIDs = null;
        
        if (xb_offerings != null && xb_offerings.length > 0) {
            for (Offering offering : xb_offerings) {
                sensorIds.add(getObservationOfferingFromOffering(offering).getProcedure());
            }
        }
        sensorIDs = sensorIds.toArray(new String[sensorIds.size()]);
        Arrays.sort(sensorIDs);
        return new SensorsGetResponse(sensorIDs);
    }

    private RestResponse handleGetSensorByIdRequest(GetSensorByIdRequest sensorsGetRequest) throws OwsExceptionReport, XmlException, IOException
    {
        SystemType xb_system = null;
        String procedureId = null;

        // 0 submit DescribeSensor (if response is an OWSException report -> cancel whole process and throw it)
        DescribeSensorRequest describeSensorRequest = sensorsGetRequest.getDescribeSensorRequest();
        procedureId = describeSensorRequest.getProcedure();
        ServiceOperator serviceOperator = getServiceOperatorFor(describeSensorRequest);
        ServiceResponse describeSensorResponse = null;
        try
        {
            describeSensorResponse = serviceOperator.receiveRequest(describeSensorRequest);
        }
        catch (OwsExceptionReport owsER) {
            if (!owsER.getExceptions().isEmpty())
            {
                for (CodedException owsE : owsER.getExceptions())
                {
                    if (owsE.getCode().equals(OwsExceptionCode.InvalidParameterValue) &&
                            owsE.getLocator().equals(SosConstants.DescribeSensorParams.procedure.toString()))
                    {
                        return new ResourceNotFoundResponse(bindingConstants.getResourceSensors(), procedureId);
                    }
                }
            }
            throw owsER;
        }

        String describeSensor = new String(describeSensorResponse.getByteArray());
        XmlObject xb_describeSensorResponse = XmlObject.Factory.parse(describeSensor);

        if (xb_describeSensorResponse instanceof DescribeSensorResponseDocument) {

            DescribeSensorResponseDocument xb_describeSensorResponseDoc = (DescribeSensorResponseDocument) xb_describeSensorResponse;
            Description xb_descriptionDoc = xb_describeSensorResponseDoc.getDescribeSensorResponse().getDescriptionArray()[0];
            SensorDescriptionType xb_description = xb_descriptionDoc.getSensorDescription();
            xb_system = getSmlSystemFromSensorDescription(xb_description);
            
            // 1 return result 
            return new GetSensorByIdResponse(xb_system,procedureId);

        } else {
            String exceptionText = String.format("Processing of SOS core operation 'DescribeSensor' response failed. Type of could not be handled: '%s'",
                    xb_describeSensorResponse.getClass().getName());
            LOGGER.debug(exceptionText);
            throw new NoApplicableCodeException().withMessage(exceptionText);
        }
    }
    
    private SystemType getSmlSystemFromSensorDescription(SensorDescriptionType sensorDescription) throws OwsExceptionReport
    {
        try {
            SensorMLDocument xb_sensorML = SensorMLDocument.Factory.parse(sensorDescription.getData().newInputStream());

            SystemType xb_system = (SystemType) xb_sensorML.getSensorML().getMemberArray()[0].getProcess().substitute(SensorMLConstants.SYSTEM_QNAME,SystemType.type);
            return xb_system;
        } catch (IOException ioe) {
        	throw logAndCreateException(ioe,String.format("Processing of '%s' failed.",
                    GetSensorByIdResponse.class.getName()));
        } catch (XmlException xe) {
            throw logAndCreateException(xe,String.format("XML Processing of '%s' failed.",
    		        GetSensorByIdResponse.class.getName()));
        }
    }

	private CodedException logAndCreateException(Exception e, String exceptionText) throws CodedException
	{
		LOGGER.debug(exceptionText);
		return new NoApplicableCodeException().causedBy(e).withMessage("%s Exception: %s",exceptionText,e.getMessage());
	}
}
