/**
 * Copyright (C) 2013
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 */
package org.n52.sos.encode;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.opengis.gml.x32.FeaturePropertyType;
import net.opengis.om.x20.NamedValueType;
import net.opengis.om.x20.OMObservationDocument;
import net.opengis.om.x20.OMObservationType;
import net.opengis.om.x20.OMProcessPropertyType;
import net.opengis.om.x20.TimeObjectPropertyType;
import net.opengis.waterml.x20.CollectionDocument;
import net.opengis.waterml.x20.CollectionType;
import net.opengis.waterml.x20.DefaultTVPMeasurementMetadataDocument;
import net.opengis.waterml.x20.MeasureTVPType;
import net.opengis.waterml.x20.MeasurementTimeseriesDocument;
import net.opengis.waterml.x20.MeasurementTimeseriesType;
import net.opengis.waterml.x20.MonitoringPointDocument;
import net.opengis.waterml.x20.MonitoringPointType;
import net.opengis.waterml.x20.ObservationProcessDocument;
import net.opengis.waterml.x20.ObservationProcessType;
import net.opengis.waterml.x20.TVPDefaultMetadataPropertyType;
import net.opengis.waterml.x20.TVPMeasurementMetadataType;

import org.apache.xmlbeans.XmlBoolean;
import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlInteger;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlString;
import org.joda.time.DateTime;
import org.n52.sos.convert.Converter;
import org.n52.sos.convert.ConverterException;
import org.n52.sos.exception.ows.NoApplicableCodeException;
import org.n52.sos.exception.ows.concrete.DateTimeFormatException;
import org.n52.sos.exception.ows.concrete.UnsupportedEncoderInputException;
import org.n52.sos.ogc.gml.CodeType;
import org.n52.sos.ogc.gml.CodeWithAuthority;
import org.n52.sos.ogc.gml.GMLConstants;
import org.n52.sos.ogc.gml.ReferenceType;
import org.n52.sos.ogc.gml.time.ITime;
import org.n52.sos.ogc.gml.time.TimeInstant;
import org.n52.sos.ogc.gml.time.TimePeriod;
import org.n52.sos.ogc.om.AbstractSosPhenomenon;
import org.n52.sos.ogc.om.NamedValue;
import org.n52.sos.ogc.om.OMConstants;
import org.n52.sos.ogc.om.SosCompositePhenomenon;
import org.n52.sos.ogc.om.SosMultiObservationValues;
import org.n52.sos.ogc.om.SosObservableProperty;
import org.n52.sos.ogc.om.SosObservation;
import org.n52.sos.ogc.om.SosSingleObservationValue;
import org.n52.sos.ogc.om.TimeValuePair;
import org.n52.sos.ogc.om.features.SosAbstractFeature;
import org.n52.sos.ogc.om.features.SosFeatureCollection;
import org.n52.sos.ogc.om.features.samplingFeatures.SosSamplingFeature;
import org.n52.sos.ogc.om.values.BooleanValue;
import org.n52.sos.ogc.om.values.CategoryValue;
import org.n52.sos.ogc.om.values.CountValue;
import org.n52.sos.ogc.om.values.GeometryValue;
import org.n52.sos.ogc.om.values.IValue;
import org.n52.sos.ogc.om.values.QuantityValue;
import org.n52.sos.ogc.om.values.TVPValue;
import org.n52.sos.ogc.om.values.TextValue;
import org.n52.sos.ogc.ows.OwsExceptionReport;
import org.n52.sos.ogc.sos.Sos2Constants;
import org.n52.sos.ogc.sos.SosConstants;
import org.n52.sos.ogc.sos.SosConstants.HelperValues;
import org.n52.sos.ogc.sos.SosProcedureDescription;
import org.n52.sos.ogc.wml.ConformanceClassesWML2;
import org.n52.sos.ogc.wml.ObservationProcess;
import org.n52.sos.ogc.wml.WaterMLConstants;
import org.n52.sos.response.GetObservationResponse;
import org.n52.sos.service.Configurator;
import org.n52.sos.service.ServiceConstants.SupportedTypeKey;
import org.n52.sos.service.profile.Profile;
import org.n52.sos.util.CodingHelper;
import org.n52.sos.util.CollectionHelper;
import org.n52.sos.util.DateTimeHelper;
import org.n52.sos.util.GmlHelper;
import org.n52.sos.util.JavaHelper;
import org.n52.sos.util.OMHelper;
import org.n52.sos.util.SosHelper;
import org.n52.sos.util.StringHelper;
import org.n52.sos.util.XmlHelper;
import org.n52.sos.util.XmlOptionsHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.vividsolutions.jts.geom.Geometry;

/**
 * Encoder class for WaterML 2.0
 * 
 */
public class WmlEncoderv20 implements ObservationEncoder<XmlObject, Object> {

    private static final Logger LOGGER = LoggerFactory.getLogger(WmlEncoderv20.class);

    private static final String DEFAULT_FEATUREOFINTEREST_ENCODING_NAMESPACE = WaterMLConstants.NS_WML2;

    // TODO: change to correct conformance class
    private static final Set<String> CONFORMANCE_CLASSES = CollectionHelper.set(
            ConformanceClassesWML2.UML_MEASUREMENT_TIMESERIES_TVP_OBSERVATION,
            ConformanceClassesWML2.UML_TIMESERIES_TVP_OBSERVATION,
            ConformanceClassesWML2.UML_MEASUREMENT_TIMESERIES_TVP_OBSERVATION, ConformanceClassesWML2.XSD_XML_RULES,
            ConformanceClassesWML2.XSD_TIMESERIES_OBSERVATION, ConformanceClassesWML2.XSD_TIMESERIES_TVP_OBSERVATION,
            ConformanceClassesWML2.XSD_MEASUREMENT_TIMESERIES_TVP);

    private static final Set<EncoderKey> ENCODER_KEYS = CodingHelper.encoderKeysForElements(WaterMLConstants.NS_WML2,
            GetObservationResponse.class, SosObservation.class, SosAbstractFeature.class);

    private static final Map<SupportedTypeKey, Set<String>> SUPPORTED_TYPES = Collections.singletonMap(
            SupportedTypeKey.ObservationType, Collections.singleton(WaterMLConstants.OBSERVATION_TYPE_MEASURMENT_TVP));;

    private static final Map<String, Map<String, Set<String>>> SUPPORTED_RESPONSE_FORMATS = Collections.singletonMap(
            SosConstants.SOS,
            Collections.singletonMap(Sos2Constants.SERVICEVERSION, Collections.singleton(WaterMLConstants.NS_WML2)));

    //
    // private int sfIdCounter = 1;
    //
    // private HashMap<String, String> gmlID4sfIdentifier;
    //
    // private static final String RESPONSE_FORMAT_V1 =
    // "text/xml;subtype=\"waterml/2.0\"";
    // private static final String RESPONSE_FORMAT_V2 =
    // "http://www.opengis.net/waterml/2.0";
    //
    // private static final String contentType = "text/xml";
    @Deprecated
    private boolean supported = true;

    public WmlEncoderv20() {
        LOGGER.debug("Encoder for the following keys initialized successfully: {}!",
                StringHelper.join(", ", ENCODER_KEYS));
    }

    @Override
    public Set<EncoderKey> getEncoderKeyType() {
        return Collections.unmodifiableSet(ENCODER_KEYS);
    }

    @Override
    public Map<SupportedTypeKey, Set<String>> getSupportedTypes() {
        return Collections.unmodifiableMap(SUPPORTED_TYPES);
    }

    @Override
    public Set<String> getConformanceClasses() {
        return Collections.unmodifiableSet(CONFORMANCE_CLASSES);
    }

    @Override
    public void addNamespacePrefixToMap(Map<String, String> nameSpacePrefixMap) {
        nameSpacePrefixMap.put(WaterMLConstants.NS_WML2, WaterMLConstants.NS_WML2_PREFIX);
    }

    @Override
    public Set<String> getSupportedResponseFormats(String service, String version) {
        if (supported && SUPPORTED_RESPONSE_FORMATS.get(service) != null) {
            if (SUPPORTED_RESPONSE_FORMATS.get(service).get(version) != null) {
                return SUPPORTED_RESPONSE_FORMATS.get(service).get(version);
            }
        }
        return Collections.emptySet();
    }

    @Override
    public boolean isObservationAndMeasurmentV20Type() {
        return true;
    }

    @Override
    @Deprecated
    public boolean isSupported() {
        return supported;
    }

    @Override
    @Deprecated
    public void setSupported(boolean supportted) {
        this.supported = supportted;
    }

    @Override
    public boolean shouldObservationsWithSameXBeMerged() {
        return true;
    }

    @Override
    public String getContentType() {
        return WaterMLConstants.WML_CONTENT_TYPE;
    }

    @Override
    public XmlObject encode(Object element) throws OwsExceptionReport {
        return encode(element, null);
    }

    @Override
    public XmlObject encode(Object element, Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        if (element instanceof GetObservationResponse) {
            return createWmlGetObservationResponse((GetObservationResponse) element);
        } else if (element instanceof SosObservation) {
            return createWaterMLObservation((SosObservation) element, additionalValues);
        } else if (element instanceof SosAbstractFeature) {
            return createMonitoringPoint((SosAbstractFeature) element);
        }
        throw new UnsupportedEncoderInputException(this, element);
    }

    private XmlObject createWmlGetObservationResponse(GetObservationResponse getObservationResonse)
            throws OwsExceptionReport {
        // TODO: set schemaLocation if final
        Map<String, String> gmlID4sfIdentifier = CollectionHelper.map();
        int sfIdCounter = 1;
        if (getObservationResonse.getObservationCollection() != null
                && !getObservationResonse.getObservationCollection().isEmpty()) {
            Collection<SosObservation> sosObservations = getObservationResonse.getObservationCollection();
            if (sosObservations.size() == 1) {
                OMObservationDocument omObservationDoc =
                        OMObservationDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
                for (SosObservation sosObservation : sosObservations) {
                    Map<HelperValues, String> foiHelper =
                            new EnumMap<SosConstants.HelperValues, String>(SosConstants.HelperValues.class);
                    String gmlId = "sf_" + sfIdCounter;
                    foiHelper.put(HelperValues.GMLID, gmlId);
                    omObservationDoc.setOMObservation((OMObservationType) createWaterMLObservation(sosObservation,
                            foiHelper));
                }
                return omObservationDoc;
            } else {
                CollectionDocument xmlCollectionDoc =
                        CollectionDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
                CollectionType wmlCollection = xmlCollectionDoc.addNewCollection();
                for (SosObservation sosObservation : sosObservations) {
                    Map<HelperValues, String> foiHelper =
                            new EnumMap<SosConstants.HelperValues, String>(SosConstants.HelperValues.class);
                    String gmlId;
                    // FIXME CodeWithAuthority VS. String keys
                    if (gmlID4sfIdentifier.containsKey(sosObservation.getObservationConstellation()
                            .getFeatureOfInterest().getIdentifier())) {
                        gmlId =
                                gmlID4sfIdentifier.get(sosObservation.getObservationConstellation()
                                        .getFeatureOfInterest().getIdentifier());
                        foiHelper.put(HelperValues.EXIST_FOI_IN_DOC, Boolean.toString(true));
                    } else {
                        gmlId = "sf_" + sfIdCounter;
                        gmlID4sfIdentifier.put(sosObservation.getObservationConstellation().getFeatureOfInterest()
                                .getIdentifier().getValue(), gmlId);
                        foiHelper.put(HelperValues.EXIST_FOI_IN_DOC, Boolean.toString(false));
                    }
                    foiHelper.put(HelperValues.GMLID, gmlId);
                    wmlCollection.addNewObservationMember().setOMObservation(
                            (OMObservationType) createWaterMLObservation(sosObservation, foiHelper));
                }
                return xmlCollectionDoc;
            }

        } else {
            // TODO: HydrologieProfile-Exception
            throw new NoApplicableCodeException().withMessage("Combination does not exists!");
        }
    }

    private XmlObject createWaterMLObservation(SosObservation sosObservation,
                                               Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        OMObservationType xbObs =
                OMObservationType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        xbObs.setId("o_" + Long.toString(System.currentTimeMillis()));
        String observationID;
        if (sosObservation.isSetObservationID()) {
            observationID = sosObservation.getObservationID();
        } else {
            observationID = xbObs.getId().replace("o_", "");
            sosObservation.setObservationID(observationID);
        }
        if (sosObservation.isSetIdentifier()) {
            Encoder<?, CodeWithAuthority> encoder =
                    Configurator
                            .getInstance()
                            .getCodingRepository()
                            .getEncoder(
                                    CodingHelper.getEncoderKey(GMLConstants.NS_GML_32, sosObservation.getIdentifier()));
            if (encoder != null) {
                XmlObject xmlObject = (XmlObject) encoder.encode(sosObservation.getIdentifier());
                xbObs.addNewIdentifier().set(xmlObject);
            } else {
                throw new NoApplicableCodeException()
                        .withMessage("Error while encoding geometry value, needed encoder is missing!");
            }
        }

        // TODO set observationType depending on result encoding
//        String observationType = sosObservation.getObservationConstellation().getObservationType();
        xbObs.addNewType().setHref(WaterMLConstants.OBSERVATION_TYPE_MEASURMENT_TVP);
        
        // set phenomenonTime
        ITime phenomenonTime = sosObservation.getPhenomenonTime();
        if (phenomenonTime.getGmlId() == null) {
            phenomenonTime.setGmlId(OMConstants.PHENOMENON_TIME_NAME + "_" + observationID);
        }
        addPhenomenonTime(xbObs.addNewPhenomenonTime(), phenomenonTime);

        // set resultTime
        addResultTime(xbObs, sosObservation);

        // set procedure
        addProcedure(xbObs.addNewProcedure(), sosObservation.getObservationConstellation().getProcedure(),
                observationID);
        // set observedProperty (phenomenon)
        List<SosObservableProperty> phenComponents;
        if (sosObservation.getObservationConstellation().getObservableProperty() instanceof SosObservableProperty) {
            xbObs.addNewObservedProperty().setHref(
                    sosObservation.getObservationConstellation().getObservableProperty().getIdentifier());
            phenComponents = new ArrayList<SosObservableProperty>(1);
            phenComponents.add((SosObservableProperty) sosObservation.getObservationConstellation()
                    .getObservableProperty());
        } else if (sosObservation.getObservationConstellation().getObservableProperty() instanceof SosCompositePhenomenon) {
            SosCompositePhenomenon compPhen =
                    (SosCompositePhenomenon) sosObservation.getObservationConstellation().getObservableProperty();
            xbObs.addNewObservedProperty().setHref(compPhen.getIdentifier());
            phenComponents = compPhen.getPhenomenonComponents();
        }
        // set feature
        encodeFeatureOfInterest(xbObs, sosObservation.getObservationConstellation().getFeatureOfInterest());
        // set result
        // TODO check for measurement values and categorical values.
        xbObs.addNewResult().set(createMeasurementTimeseries(sosObservation, phenomenonTime.getGmlId()));
        return xbObs;
    }

    private ObservationProcess convertProcedureDescription(SosProcedureDescription procedureDescription)
            throws OwsExceptionReport {
        Converter<?, SosProcedureDescription> converter =
                Configurator.getInstance().getConverterRepository()
                        .getConverter(procedureDescription.getDescriptionFormat(), WaterMLConstants.NS_WML2);
        if (converter != null) {
            try {
                Object object = converter.convert(procedureDescription);
                if (object instanceof ObservationProcess) {
                    return (ObservationProcess) object;
                }
            } catch (ConverterException e) {
                throw new NoApplicableCodeException().withMessage("Error while converting procedureDescription!");
            }
        }
        return null;
    }

    private void addProcedure(OMProcessPropertyType procedure, SosProcedureDescription procedureDescription,
                              String observationID) throws OwsExceptionReport {
        if (Configurator.getInstance().getProfileHandler().getActiveProfile()
                .isEncodeProcedureInObservation(WaterMLConstants.NS_WML2)) {
            if (procedureDescription instanceof ObservationProcess) {
                procedure.set(createObservationProcess((ObservationProcess) procedureDescription, observationID));
            } else {
                ObservationProcess convertedDescription = convertProcedureDescription(procedureDescription);
                if (convertedDescription != null) {
                    procedure.set(createObservationProcess(convertedDescription, observationID));
                } else {
                    procedure.setHref(procedureDescription.getIdentifier());
                }
            }
        } else {
            procedure.setHref(procedureDescription.getIdentifier());
        }
    }

    private ObservationProcessDocument createObservationProcess(ObservationProcess procedure, String gmlID)
            throws OwsExceptionReport {
        ObservationProcessDocument observationProcessDoc = ObservationProcessDocument.Factory.newInstance();
        ObservationProcessType observationProcess = observationProcessDoc.addNewObservationProcess();
        observationProcess.setId("process." + gmlID);
        addProcessType(observationProcess, procedure);
        addOriginatingProcess(observationProcess, procedure);
        addAggregatingDuration(observationProcess, procedure);
        addVerticalDatum(observationProcess, procedure);
        addComment(observationProcess, procedure);
        addProcessReference(observationProcess, procedure);
        addInput(observationProcess, procedure);
        addParameter(observationProcess, procedure);
        return observationProcessDoc;
    }

    private void addProcessType(ObservationProcessType observationProcess, ObservationProcess procedure)
            throws OwsExceptionReport {
        if (procedure.isSetProcessType() && procedure.getProcessType().isSetHref()) {
            XmlObject referenceType = encodeReferenceType(procedure.getProcessType());
            if (referenceType != null) {
                observationProcess.addNewProcessType().set(referenceType);
            }
        } else {
            throw new NoApplicableCodeException().withMessage("Missing processType definition");
        }
    }

    private void addOriginatingProcess(ObservationProcessType observationProcess, ObservationProcess procedure)
            throws OwsExceptionReport {
        if (procedure.isSetOriginatingProcess()) {
            XmlObject referenceType = encodeReferenceType(procedure.getOriginatingProcess());
            if (referenceType != null) {
                observationProcess.addNewOriginatingProcess().set(referenceType);
            }
        }
    }

    private void addAggregatingDuration(ObservationProcessType observationProcess, ObservationProcess procedure) {
    }

    private void addVerticalDatum(ObservationProcessType observationProcess, ObservationProcess procedure)
            throws OwsExceptionReport {
        if (procedure.isSetVerticalDatum()) {
            XmlObject referenceType = encodeReferenceType(procedure.getVerticalDatum());
            if (referenceType != null) {
                observationProcess.addNewVerticalDatum().set(referenceType);
            }
        }
    }

    private void addComment(ObservationProcessType observationProcess, ObservationProcess procedure) {
        if (procedure.isSetComments()) {
            for (String comment : procedure.getComments()) {
                if (comment != null && !comment.isEmpty()) {
                    observationProcess.addComment(comment);
                }
            }
        }
    }

    private void addProcessReference(ObservationProcessType observationProcess, ObservationProcess procedure)
            throws OwsExceptionReport {
        if (procedure.isSetProcessReference()) {
            XmlObject referenceType = encodeReferenceType(procedure.getProcessReference());
            if (referenceType != null) {
                observationProcess.addNewProcessReference().set(referenceType);
            }
        }
    }

    private void addInput(ObservationProcessType observationProcess, ObservationProcess procedure)
            throws OwsExceptionReport {
        if (procedure.isSetInputs()) {
            for (org.n52.sos.ogc.gml.ReferenceType sosReferenceType : procedure.getInputs()) {
                XmlObject referenceType = encodeReferenceType(sosReferenceType);
                if (referenceType != null) {
                    observationProcess.addNewInput().set(referenceType);
                }
            }
        }
    }

    private void addParameter(ObservationProcessType observationProcess, ObservationProcess procedure)
            throws OwsExceptionReport {
        if (procedure.isSetParameters()) {
            List<NamedValue<?>> parameters = procedure.getParameters();
            for (NamedValue<?> sosNamedValue : parameters) {
                XmlObject namedValue = createNamedValue(sosNamedValue);
                if (namedValue != null) {
                    observationProcess.addNewParameter().addNewNamedValue().set(namedValue);
                }
            }
        }
    }
    
    private XmlObject createNamedValue(NamedValue<?> sosNamedValue) throws OwsExceptionReport {
        // encode value (any)
        XmlObject namedValuePropertyValue = getNamedValueValue(sosNamedValue.getValue());
        if (namedValuePropertyValue != null) {
            NamedValueType xbNamedValue =
                    NamedValueType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
            // encode gml:ReferenceType
            XmlObject encodeObjectToXml =
                    CodingHelper.encodeObjectToXml(GMLConstants.NS_GML_32, sosNamedValue.getName());
            xbNamedValue.addNewName().set(encodeObjectToXml);
            // set value (any)
            xbNamedValue.setValue(namedValuePropertyValue);
            return xbNamedValue;
        }
        return null;
    }

    private XmlObject getNamedValueValue(IValue<?> iValue) throws OwsExceptionReport {
        if (iValue instanceof BooleanValue) {
            BooleanValue booleanValue = (BooleanValue) iValue;
            XmlBoolean xbBoolean = XmlBoolean.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
            if (booleanValue.getValue() != null) {
                xbBoolean.setBooleanValue(booleanValue.getValue());
                return xbBoolean;
            }
        } else if (iValue instanceof CategoryValue) {
            CategoryValue categoryValue = (CategoryValue) iValue;
            if (categoryValue.getValue() != null && !categoryValue.getValue().isEmpty()) {
                return CodingHelper.encodeObjectToXml(GMLConstants.NS_GML_32, categoryValue);
            }
        } else if (iValue instanceof CountValue) {
            CountValue countValue = (CountValue) iValue;
            XmlInteger xbInteger = XmlInteger.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
            if (countValue.getValue() != null && countValue.getValue() != Integer.MIN_VALUE) {
                xbInteger.setBigIntegerValue(new BigInteger(countValue.getValue().toString()));
                return xbInteger;
            }
        } else if (iValue instanceof GeometryValue) {
            GeometryValue geometryValue = (GeometryValue) iValue;
            if (geometryValue.getValue() != null) {
                return CodingHelper.encodeObjectToXml(GMLConstants.NS_GML_32, geometryValue.getValue());
            }
        } else if (iValue instanceof QuantityValue) {
            QuantityValue quantityValue = (QuantityValue) iValue;
            return CodingHelper.encodeObjectToXml(GMLConstants.NS_GML_32, quantityValue);
        } else if (iValue instanceof TextValue) {
            TextValue textValue = (TextValue) iValue;
            XmlString xbString = XmlString.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
            xbString.setStringValue(textValue.getValue());
            return xbString;
        }
        return null;
    }

    private XmlObject encodeReferenceType(ReferenceType sosReferenceType) throws OwsExceptionReport {
        Encoder<XmlObject, ReferenceType> encoder =
                Configurator.getInstance().getCodingRepository()
                        .getEncoder(CodingHelper.getEncoderKey(GMLConstants.NS_GML_32, sosReferenceType));
        if (encoder != null) {
            return encoder.encode(sosReferenceType);
        } else {
            throw new NoApplicableCodeException()
                    .withMessage("Error while encoding referenceType, needed encoder is missing!");
        }

    }

    private XmlObject createMonitoringPoint(SosAbstractFeature absFeature) throws OwsExceptionReport {
        SosSamplingFeature sampFeat = (SosSamplingFeature) absFeature;
        if (absFeature.isSetGmlID()) {
            FeaturePropertyType featureProperty = FeaturePropertyType.Factory.newInstance();
            featureProperty.setHref("#" + absFeature.getGmlId());
            return featureProperty;
        } else {
            StringBuilder builder = new StringBuilder();
            builder.append("mp_");
            builder.append(JavaHelper.generateID(absFeature.getIdentifier().getValue()));
            absFeature.setGmlId(builder.toString());
            MonitoringPointDocument monitoringPointDoc = MonitoringPointDocument.Factory.newInstance();
            if (sampFeat.getXmlDescription() != null) {
                try {
                    XmlObject feature = XmlObject.Factory.parse(sampFeat.getXmlDescription());
                    if (XmlHelper.getNamespace(feature).equals(WaterMLConstants.NS_WML2)) {
                        if (feature instanceof MonitoringPointDocument) {
                            monitoringPointDoc = (MonitoringPointDocument) feature;
                        } else if (feature instanceof MonitoringPointType) {
                            monitoringPointDoc.setSFSpatialSamplingFeature((MonitoringPointType) feature);
                        }
                        XmlHelper.updateGmlIDs(monitoringPointDoc.getDomNode(), absFeature.getGmlId(), null);
                        return monitoringPointDoc;
                    }
                } catch (XmlException xmle) {
                    throw new NoApplicableCodeException().causedBy(xmle)
                            .withMessage("Error while encoding GetFeatureOfInterest response, invalid samplingFeature description!");
                }
            }
            MonitoringPointType monitoringPoint = monitoringPointDoc.addNewMonitoringPoint();
            // set gml:id
            monitoringPoint.setId(absFeature.getGmlId());

            if (sampFeat.isSetIdentifier()
                    && SosHelper.checkFeatureOfInterestIdentifierForSosV2(sampFeat.getIdentifier().getValue(),
                            Sos2Constants.SERVICEVERSION)) {
                XmlObject xmlObject = CodingHelper.encodeObjectToXml(GMLConstants.NS_GML_32, sampFeat.getIdentifier());
                if (xmlObject != null) {
                    monitoringPoint.addNewIdentifier().set(xmlObject);
                }
            }

            if (sampFeat.isSetNames()) {
                for (CodeType sosName : sampFeat.getName()) {
                    XmlObject xmlObject = CodingHelper.encodeObjectToXml(GMLConstants.NS_GML_32, sosName);
                    if (xmlObject != null) {
                        monitoringPoint.addNewName().set(xmlObject);
                    }
                }
            }

            // set type
            // TODO: check if special definition
            // monitoringPoint.addNewType().setHref(sampFeat.getFeatureType());

            // set sampledFeatures
            // TODO: CHECK
            if (sampFeat.getSampledFeatures() != null && !sampFeat.getSampledFeatures().isEmpty()) {
                if (sampFeat.getSampledFeatures().size() == 1) {
                    XmlObject encodeObjectToXml = CodingHelper
                            .encodeObjectToXml(GMLConstants.NS_GML_32, sampFeat.getSampledFeatures().get(0));
                    monitoringPoint.addNewSampledFeature().set(encodeObjectToXml);
                } else {
                    SosFeatureCollection featureCollection = new SosFeatureCollection();
                    featureCollection.setGmlId("sampledFeatures_" + absFeature.getGmlId());
                    for (SosAbstractFeature sampledFeature : sampFeat.getSampledFeatures()) {
                        featureCollection.addMember(sampledFeature);
                    }
                    XmlObject encodeObjectToXml = CodingHelper
                            .encodeObjectToXml(GMLConstants.NS_GML_32, featureCollection);
                    monitoringPoint.addNewSampledFeature().set(encodeObjectToXml);
                }
            } else {
                monitoringPoint.addNewSampledFeature().setHref(GMLConstants.NIL_UNKNOWN);
            }
//            if (sampFeat.getSampledFeatures() != null && !sampFeat.getSampledFeatures().isEmpty()) {
//                for (SosAbstractFeature sampledFeature : sampFeat.getSampledFeatures()) {
//                    FeaturePropertyType sampledFeat = monitoringPoint.addNewSampledFeature();
//                    // sampledFeat.set(encodeSampledFeature((SosSamplingFeature)
//                    // sampledFeature, "xyz"));
//                    // sampledFeat.set(createMonitoringPoint(absFeature));
//                    sampledFeat.setHref(sampledFeature.getIdentifier().getValue());
//                    if (sampledFeature instanceof SosSamplingFeature) {
//                        SosSamplingFeature feat = (SosSamplingFeature) sampledFeature;
//                        if (feat.isSetNames()) {
//                            sampledFeat.setTitle(feat.getFirstName().getValue());
//                        }
//                    }
//                }
//            } else {
//                FeaturePropertyType sampledFeat = monitoringPoint.addNewSampledFeature();
//                sampledFeat.setHref(GMLConstants.NIL_UNKNOWN);
//            }
            
            if (sampFeat.isSetParameter()) {
                addParameter(monitoringPoint, sampFeat);
            }

            // set position
            Encoder<XmlObject, Geometry> encoder =
                    Configurator.getInstance().getCodingRepository()
                            .getEncoder(CodingHelper.getEncoderKey(GMLConstants.NS_GML_32, sampFeat.getGeometry()));
            if (encoder != null) {
                Map<HelperValues, String> additionalValues = new EnumMap<HelperValues, String>(HelperValues.class);
                additionalValues.put(HelperValues.GMLID, absFeature.getGmlId());
                XmlObject xmlObject = encoder.encode(sampFeat.getGeometry(), additionalValues);
                monitoringPoint.addNewShape().addNewAbstractGeometry().set(xmlObject);
                XmlHelper.substituteElement(monitoringPoint.getShape().getAbstractGeometry(), xmlObject);
            } else {
                throw new NoApplicableCodeException()
                        .withMessage("Error while encoding geometry for feature, needed encoder is missing!");
            }
            return monitoringPointDoc;
        }
    }
    
    private void addParameter(MonitoringPointType monitoringPoint, SosSamplingFeature sampFeat) throws OwsExceptionReport {
        for (NamedValue<?> namedValue : sampFeat.getParameters()) {
            XmlObject encodeObjectToXml = createNamedValue(namedValue);
            if (encodeObjectToXml != null) {
                monitoringPoint.addNewParameter().addNewNamedValue().set(encodeObjectToXml);
            }
        }
    }

    private XmlObject createMeasurementTimeseries(SosObservation sosObservation, String phenomenonTimeID)
            throws OwsExceptionReport {
        MeasurementTimeseriesDocument measurementTimeseriesDoc = MeasurementTimeseriesDocument.Factory.newInstance();
        MeasurementTimeseriesType measurementTimeseries = measurementTimeseriesDoc.addNewMeasurementTimeseries();
        measurementTimeseries.setId("timeseries." + sosObservation.getObservationID());
        measurementTimeseries.addNewMetadata().addNewTimeseriesMetadata().addNewTemporalExtent()
                .setHref("#" + phenomenonTimeID);

        TVPDefaultMetadataPropertyType xbMetaComponent = measurementTimeseries.addNewDefaultPointMetadata();

        DefaultTVPMeasurementMetadataDocument xbDefMeasureMetaComponent =
                DefaultTVPMeasurementMetadataDocument.Factory.newInstance();
        TVPMeasurementMetadataType defaultTVPMeasurementMetadata =
                xbDefMeasureMetaComponent.addNewDefaultTVPMeasurementMetadata();
        defaultTVPMeasurementMetadata.addNewInterpolationType().setHref(
                "http://www.opengis.net/def/timeseriesType/WaterML/2.0/continuous");

        xbDefMeasureMetaComponent.getDefaultTVPMeasurementMetadata().getInterpolationType().setTitle("Instantaneous");
        String unit = null;
        if (sosObservation.getValue() instanceof SosSingleObservationValue) {
            SosSingleObservationValue<?> singleObservationValue =
                    (SosSingleObservationValue) sosObservation.getValue();
            String time = getTimeString(singleObservationValue.getPhenomenonTime());
            unit = singleObservationValue.getValue().getUnit();
            if (sosObservation.getValue().getValue() instanceof QuantityValue) {
                QuantityValue quantityValue = (QuantityValue) singleObservationValue.getValue();
                //FIXME equals on BigDecimal and Double.NaN
                if (!quantityValue.getValue().equals(Double.NaN)) {
                    String value = Double.toString(quantityValue.getValue().doubleValue());
                    addValuesToMeasurementTVP(measurementTimeseries.addNewPoint().addNewMeasurementTVP(), time,
                            value);
                }
            } else if (sosObservation.getValue().getValue() instanceof CountValue) {
                CountValue countValue = (CountValue) singleObservationValue.getValue();
                if (countValue.getValue() != null) {
                    String value = Integer.toString(countValue.getValue().intValue());
                    addValuesToMeasurementTVP(measurementTimeseries.addNewPoint().addNewMeasurementTVP(), time,
                            value);
                }
            }
        } else if (sosObservation.getValue() instanceof SosMultiObservationValues) {
            SosMultiObservationValues<?> observationValue = (SosMultiObservationValues) sosObservation.getValue();
            TVPValue tvpValue = (TVPValue) observationValue.getValue();
            List<TimeValuePair> timeValuePairs = tvpValue.getValue();
            unit = tvpValue.getUnit();
            for (TimeValuePair timeValuePair : timeValuePairs) {
                if (timeValuePair.getValue() instanceof QuantityValue) {
                    QuantityValue quantityValue = (QuantityValue) timeValuePair.getValue();
                    //FIXME equals on BigDecimal and Double.NaN
                    if (!quantityValue.getValue().equals(Double.NaN)) {
                        timeValuePair.getTime();
                        String time = getTimeString(timeValuePair.getTime());
                        String value = Double.toString(quantityValue.getValue().doubleValue());
                        addValuesToMeasurementTVP(measurementTimeseries.addNewPoint().addNewMeasurementTVP(),
                                time, value);
                    }
                } else if (timeValuePair.getValue() instanceof CountValue) {
                    CountValue countValue = (CountValue) timeValuePair.getValue();
                    if (countValue.getValue() != null) {
                        String time = getTimeString(timeValuePair.getTime());
                        String value = Integer.toString(countValue.getValue().intValue());
                        addValuesToMeasurementTVP(measurementTimeseries.addNewPoint().addNewMeasurementTVP(),
                                time, value);
                    }
                }
            }
        }
        // set uom
        if (unit != null && !unit.isEmpty()) {
            defaultTVPMeasurementMetadata.addNewUom().setCode(unit);
        } else {
            SosObservableProperty observableProperty = (SosObservableProperty)sosObservation.getObservationConstellation().getObservableProperty();
            if (observableProperty.isSetUnit()) {
                defaultTVPMeasurementMetadata.addNewUom().setCode(observableProperty.getUnit());
            } 
        }
        
        xbMetaComponent.set(xbDefMeasureMetaComponent);
        return measurementTimeseriesDoc;
    }

    private void addValuesToMeasurementTVP(MeasureTVPType measurementTVP, String time, String value) {
        measurementTVP.addNewTime().setStringValue(time);
        if (value != null && !value.isEmpty()) {
            measurementTVP.addNewValue().setStringValue(value);
        } else {
            measurementTVP.addNewValue().setNil();
            measurementTVP.addNewMetadata().addNewTVPMeasurementMetadata().addNewNilReason().setNilReason("missing");
        }
    }

    private String getTimeString(ITime iTtime) throws DateTimeFormatException {
        DateTime time = getTime(iTtime);
        return DateTimeHelper.formatDateTime2String(time, iTtime.getTimeFormat());
    }

    private DateTime getTime(ITime time) {
        if (time instanceof TimeInstant) {
            return ((TimeInstant) time).getValue();
        } else if (time instanceof TimePeriod) {
            TimePeriod timePeriod = (TimePeriod) time;
            if (timePeriod.getEnd() != null) {
                return timePeriod.getEnd();
            } else {
                return timePeriod.getStart();
            }
        }
        return new DateTime().minusYears(1000);
    }

    /**
     * Encodes a SosAbstractFeature to an SpatialSamplingFeature under
     * consideration of duplicated SpatialSamplingFeature in the XML document.
     * 
     * @param xbObs
     *            XmlObject O&M observation
     * @param absObs
     *            SOS observation

     *
     * @throws OwsExceptionReport
     */
    private void encodeFeatureOfInterest(OMObservationType observation, SosAbstractFeature feature)
            throws OwsExceptionReport {

        SosSamplingFeature samplingFeature = (SosSamplingFeature) feature;
        FeaturePropertyType featureProperty = observation.addNewFeatureOfInterest();
        Profile activeProfile = Configurator.getInstance().getProfileHandler().getActiveProfile();
        if (!activeProfile.isEncodeFeatureOfInterestInObservations() || !(feature instanceof SosSamplingFeature)) {
            featureProperty.setHref(feature.getIdentifier().getValue());
            if (samplingFeature.isSetNames()) {
                featureProperty.setTitle(samplingFeature.getFirstName().getValue());
            }
        } else {
            if (samplingFeature.getUrl() != null) {
                featureProperty.setHref(samplingFeature.getUrl());
            } else {
                if (activeProfile.isSetEncodeFeatureOfInterestNamespace()
                        && !activeProfile.getEncodingNamespaceForFeatureOfInterest().equals(
                                DEFAULT_FEATUREOFINTEREST_ENCODING_NAMESPACE)) {
                    XmlObject encodedXmlObject =
                            CodingHelper.encodeObjectToXml(
                                    OMHelper.getNamespaceForFeatureType(samplingFeature.getFeatureType()),
                                    samplingFeature);

                    if (encodedXmlObject != null) {
                        featureProperty.set(encodedXmlObject);
                    } else {
                        if (samplingFeature.getXmlDescription() != null) {
                            try {
                                // TODO how set gml:id in already existing
                                // XmlDescription?
                                featureProperty.set(XmlObject.Factory.parse(samplingFeature.getXmlDescription()));
                            } catch (XmlException xmle) {
                                throw new NoApplicableCodeException().causedBy(xmle)
                                        .withMessage("Error while encoding featureOfInterest in OMObservation!");
                            }
                        } else {
                            featureProperty.setHref(samplingFeature.getIdentifier().getValue());
                            if (samplingFeature.isSetNames()) {
                                featureProperty.setTitle(samplingFeature.getFirstName().getValue());
                            }
                        }
                    }
                } else {
                    featureProperty.set(createMonitoringPoint(feature));
                }
            }
        }
    }

    private void addPhenomenonTime(TimeObjectPropertyType timeObjectPropertyType, ITime iTime)
            throws OwsExceptionReport {
        Encoder<XmlObject, ITime> encoder =
                                  Configurator.getInstance().getCodingRepository()
                        .getEncoder(CodingHelper.getEncoderKey(GMLConstants.NS_GML_32, iTime));
        if (encoder != null) {
            XmlObject xmlObject = encoder.encode(iTime);
            XmlObject substitution =
                    timeObjectPropertyType.addNewAbstractTimeObject().substitute(GmlHelper.getQnameForITime(iTime),
                            xmlObject.schemaType());
            substitution.set(xmlObject);
        } else {
            throw new NoApplicableCodeException()
                    .withMessage("Error while encoding phenomenon time, needed encoder is missing!");
        }
    }

    private void addResultTime(OMObservationType xbObs, SosObservation sosObservation) throws OwsExceptionReport {
        TimeInstant resultTime = sosObservation.getResultTime();
        ITime phenomenonTime = sosObservation.getPhenomenonTime();
        if (sosObservation.getResultTime() != null) {
            if (resultTime.equals(phenomenonTime)) {
                xbObs.addNewResultTime().setHref("#" + phenomenonTime.getGmlId());
            } else {
                addResultTime(xbObs, resultTime);
            }
        } else {
            if (phenomenonTime instanceof TimeInstant) {
                xbObs.addNewResultTime().setHref("#" + phenomenonTime.getGmlId());
            } else if (phenomenonTime instanceof TimePeriod) {
                TimePeriod phenTime = (TimePeriod) sosObservation.getPhenomenonTime();
                TimeInstant rsTime = new TimeInstant(phenTime.getEnd());
                rsTime.setTimeFormat(phenTime.getTimeFormat());
                addResultTime(xbObs, rsTime);
            }
        }
    }

    private void addResultTime(OMObservationType xbObs, TimeInstant iTime) throws OwsExceptionReport {
        Encoder<XmlObject, TimeInstant> encoder = CodingHelper.getEncoder(GMLConstants.NS_GML_32, iTime);
        if (encoder != null) {
            XmlObject xmlObject = encoder.encode(iTime);
            xbObs.addNewResultTime().addNewTimeInstant().set(xmlObject);
            XmlObject substitution =
                    xbObs.getResultTime().getTimeInstant()
                            .substitute(GmlHelper.getQnameForITime(iTime), xmlObject.schemaType());
            substitution.set(xmlObject);
        } else {
            throw new NoApplicableCodeException()
                    .withMessage("Error while encoding phenomenon time, needed encoder is missing!");
        }
    }

    private XmlObject encodeSampledFeature(SosSamplingFeature sampFeat, String gmlId) throws OwsExceptionReport {
        Encoder<XmlObject, SosSamplingFeature> encoder =
                Configurator
                        .getInstance()
                        .getCodingRepository()
                        .getEncoder(
                                new XmlEncoderKey(OMHelper.getNamespaceForFeatureType(sampFeat.getFeatureType()),
                                        SosSamplingFeature.class));
        if (encoder != null) {
            Map<HelperValues, String> additionalValues = new EnumMap<HelperValues, String>(HelperValues.class);
            additionalValues.put(HelperValues.GMLID, gmlId);
            return encoder.encode(sampFeat, additionalValues);
        } else {
            throw new NoApplicableCodeException()
                    .withMessage("Error while encoding geometry for featureOfInterest, needed encoder is missing!");
        }
    }

}