/**
 * Copyright (C) 2013
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 */
package org.n52.sos.ogc.wml;

import java.util.ArrayList;
import java.util.List;

import org.n52.sos.ogc.gml.ReferenceType;
import org.n52.sos.ogc.om.NamedValue;
import org.n52.sos.ogc.om.SosOffering;
import org.n52.sos.ogc.sos.SosProcedureDescription;

public class ObservationProcess extends SosProcedureDescription {

    private String procedureIdentifier;

    /*
     * Multiplicity: 1 
     * A defintion of the type of process used in the
     * observation. This may be a Sensor, ManualMethod, Algorithm or Simulation
     * (including models).
     */
    private ReferenceType processType;

    /*
     * Multiplicity: 0..1 
     * A reference to the original source of the data. For
     * example, if this is a post-processed time series (and processType is
     * algorithm), this link would specify the original process that generated
     * the data, e.g. the sensor. This allows the origin of the data to be
     * maintained regardless of the processing that has occured to it.
     */
    private ReferenceType originatingProcess;

    /*
     * Multiplicity: 0..1 
     * If the process involves temporal aggregation of a
     * result set, the time duration over which data has been aggregated should
     * be expressed here. E.g. hourly, daily aggregates.
     */
    private String aggregationDuration;

    /*
     * Multiplicity: 0..1 
     * Reference to an external process definition
     */
    private ReferenceType processReference;

    /*
     * Multiplicity: 0..1 
     * Specifies the datum that is used as the zero point for
     * level measurements. This can be process-specific as opposed the gauge at
     * the actual monitoring point.
     */
    private ReferenceType verticalDatum;

    /*
     * Multiplicity: 0..* 
     * A list of the inputs used in the process. This may be
     * a list of references to the data sets used (e.g. model input series) or a
     * input array to an algorithm.
     */
    private List<ReferenceType> inputs = new ArrayList<ReferenceType>(0);

    /*
     * Multiplicity: 0..* 
     * Comments specific to the process from the operator or
     * system performing the process.
     */
    private List<String> comments = new ArrayList<String>(0);

    /*
     * Multiplicity: 0..* 
     * A defintion of the type of process used in the
     * observation. This may be a Sensor, ManualMethod, Algorithm or Simulation
     * (including models).
     */
    private List<NamedValue<?>> parameters = new ArrayList<NamedValue<?>>(0);

    @Override
    public String getIdentifier() {
        return procedureIdentifier;
    }

    @Override
    public List<SosOffering> getOfferingIdentifiers() {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public boolean isSetOffering() {
        // TODO Auto-generated method stub
        return false;
    }

    public void setIdentifier(String procedureIdentifier) {
        this.procedureIdentifier = procedureIdentifier;
    }

    public ReferenceType getProcessType() {
        return processType;
    }

    public void setProcessType(ReferenceType processType) {
        this.processType = processType;
    }

    public ReferenceType getOriginatingProcess() {
        return originatingProcess;
    }

    public void setOriginatingProcess(ReferenceType originatingProcess) {
        this.originatingProcess = originatingProcess;
    }

    public String getAggregationDuration() {
        return aggregationDuration;
    }

    public void setAggregationDuration(String aggregationDuration) {
        this.aggregationDuration = aggregationDuration;
    }

    public ReferenceType getProcessReference() {
        return processReference;
    }

    public void setProcessReference(ReferenceType processReference) {
        this.processReference = processReference;
    }

    public ReferenceType getVerticalDatum() {
        return verticalDatum;
    }

    public void setVerticalDatum(ReferenceType verticalDatum) {
        this.verticalDatum = verticalDatum;
    }

    public List<ReferenceType> getInputs() {
        return inputs;
    }

    public void setInputs(List<ReferenceType> input) {
        this.inputs.addAll(input);
    }

    public void addInputs(ReferenceType input) {
        this.inputs.add(input);
    }

    public List<String> getComments() {
        return comments;
    }

    public void setComments(List<String> comments) {
        this.comments = comments;
    }

    public void addComment(String comment) {
        this.comments.add(comment);
    }

    public List<NamedValue<?>> getParameters() {
        return parameters;
    }

    public void setParameters(List<NamedValue<?>> parameters) {
        this.parameters.addAll(parameters);
    }

    public void addParameter(NamedValue<?> parameter) {
        this.parameters.add(parameter);
    }
    
    public boolean isSetProcessType() {
        return processType != null;
    }
    
    public boolean isSetOriginatingProcess() {
        return originatingProcess != null && originatingProcess.hasValues();
    }
    
    public boolean isSetProcessReference() {
        return processReference != null && processReference.hasValues();
    }
    
    public boolean isSetAggregationDuration() {
        return aggregationDuration != null && !aggregationDuration.isEmpty();
    }
    
    public boolean isSetVerticalDatum() {
        return verticalDatum != null && verticalDatum.hasValues();
    }
    
    public boolean isSetInputs() {
        return inputs != null && !inputs.isEmpty();
    }
    
    public boolean isSetComments() {
        return comments != null && !comments.isEmpty();
    }
    
    public boolean isSetParameters() {
        return parameters != null && !parameters.isEmpty();
    }
    
}
