/**
 * Copyright (C) 2013
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 */
package org.n52.sos.ext.deleteobservation;

import static org.n52.sos.ext.deleteobservation.DeleteObservationConstants.*;
import static org.n52.sos.util.CodingHelper.encoderKeysForElements;

import java.util.Collections;
import java.util.Map;
import java.util.Set;

import net.opengis.sosdo.x10.DeleteObservationResponseDocument;
import net.opengis.sosdo.x10.DeleteObservationResponseType;

import org.apache.xmlbeans.XmlObject;
import org.n52.sos.encode.Encoder;
import org.n52.sos.encode.EncoderKey;
import org.n52.sos.exception.ows.NoApplicableCodeException;
import org.n52.sos.exception.ows.concrete.UnsupportedEncoderInputException;
import org.n52.sos.ogc.ows.CompositeOwsException;
import org.n52.sos.ogc.ows.OwsExceptionReport;
import org.n52.sos.ogc.sos.SosConstants.HelperValues;
import org.n52.sos.service.ServiceConstants.SupportedTypeKey;
import org.n52.sos.util.StringHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author <a href="mailto:e.h.juerrens@52north.org">Eike Hinderk J&uuml;rrens</a>
 */
public class DeleteObservationEncoder implements Encoder<XmlObject, DeleteObservationResponse> {
    
    private static final Set<EncoderKey> ENCODER_KEYS =
                                         encoderKeysForElements(NS_SOSDO_1_0, DeleteObservationResponse.class);
    
    private static final Logger LOGGER = LoggerFactory.getLogger(DeleteObservationEncoder.class);
    
    public DeleteObservationEncoder() {
    	LOGGER.info("Encoder for the following keys initialized successfully: {}!", StringHelper.join(", ", ENCODER_KEYS));
    }

    public Set<EncoderKey> getEncoderKeyType() {
        return Collections.unmodifiableSet(ENCODER_KEYS);
    }

    public Map<SupportedTypeKey, Set<String>> getSupportedTypes() {
        return Collections.emptyMap();
    }

    public Set<String> getConformanceClasses() {
        return CONFORMANCE_CLASSES;
    }
    
    public void addNamespacePrefixToMap(Map<String, String> nameSpacePrefixMap) {
    	if (nameSpacePrefixMap != null) 
    	{
    		nameSpacePrefixMap.put(NS_SOSDO_1_0, NS_SOSDO_1_0_PREFIX);
    	}
    }
    
    public String getContentType() {
        return "text/xml";
    }

    public XmlObject encode(DeleteObservationResponse dor) throws OwsExceptionReport {
        if (dor == null) {
            throw new UnsupportedEncoderInputException(this, dor);
        }
        CompositeOwsException exceptions = new CompositeOwsException();
        if (dor.getService() == null) {
            exceptions.add(new NoApplicableCodeException().withMessage("service attribute is missing"));
        }
        if (dor.getVersion() == null) {
            exceptions.add(new NoApplicableCodeException().withMessage("version attribute is missing"));
        }
        if (dor.getObservationId() == null || dor.getObservationId().isEmpty()) {
            exceptions.add(new NoApplicableCodeException().withMessage("observationId attribute is missing"));
        }
        exceptions.throwIfNotEmpty();

        String observationId = dor.getObservationId();
        DeleteObservationResponseDocument xb_DeleteObsDoc = DeleteObservationResponseDocument.Factory.newInstance();
        DeleteObservationResponseType xb_DeleteObservationResponse = xb_DeleteObsDoc.addNewDeleteObservationResponse();
        xb_DeleteObservationResponse.setDeletedObservation(observationId);
        return xb_DeleteObsDoc;
    }

    public XmlObject encode(DeleteObservationResponse serviceResponseToEncode,
                            Map<HelperValues, String> IGNOREDadditionalValues) throws OwsExceptionReport
    {
        return encode(serviceResponseToEncode);
    }


}
