/**
 * Copyright (C) 2013
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 */
package org.n52.sos.ext.deleteobservation;

import static org.hibernate.criterion.Restrictions.eq;

import org.hibernate.Criteria;
import org.hibernate.Session;
import org.hibernate.criterion.Projections;
import org.joda.time.DateTime;
import org.n52.sos.ds.hibernate.HibernateSessionHolder;
import org.n52.sos.ds.hibernate.entities.FeatureOfInterest;
import org.n52.sos.ds.hibernate.entities.Observation;
import org.n52.sos.ds.hibernate.entities.Offering;
import org.n52.sos.ds.hibernate.entities.Procedure;
import org.n52.sos.ds.hibernate.util.HibernateCriteriaQueryUtilities;
import org.n52.sos.ogc.ows.OwsExceptionReport;

/**
 * Updates the cache after a Observation was deleted. Uses the deleted observation to determine which cache relations
 * have to be updated.
 * <p/>
 * @author Christian Autermann <c.autermann@52north.org>
 * @since 4.0
 */
public class HibernateDeleteObservationCacheFeederDAO extends DeleteObservationCacheFeederDAO {
    private final HibernateSessionHolder sessionHolder = new HibernateSessionHolder();
    private Session session;

    @Override
    protected boolean isLastForProcedure(String feature, String procedure) {
        Criteria oc = getConnection().createCriteria(Observation.class)
                .add(eq(Observation.DELETED, false));
        oc.createCriteria(Observation.FEATURE_OF_INTEREST)
                .add(eq(FeatureOfInterest.IDENTIFIER, feature));
        oc.createCriteria(Observation.PROCEDURE)
                .add(eq(Procedure.IDENTIFIER, procedure));
        return isEmpty(oc);
    }

    @Override
    protected boolean isLastForOffering(String feature, String offering) {
        Criteria oc = getConnection().createCriteria(Observation.class)
                .add(eq(Observation.DELETED, false));
        oc.createCriteria(Observation.FEATURE_OF_INTEREST)
                .add(eq(FeatureOfInterest.IDENTIFIER, feature));
        oc.createCriteria(Observation.OFFERINGS)
                .add(eq(Offering.IDENTIFIER, offering));
        return isEmpty(oc);
    }

    /**
     * Checks if the specified query has no results.
     *
     * @param q the query
     *
     * @return if it has no results
     */
    protected boolean isEmpty(Criteria q) {
        return ((Number) q.setProjection(Projections.rowCount()).uniqueResult()).longValue() == 0L;
    }

    @Override
    protected DateTime getMaxResultTime() {
        return HibernateCriteriaQueryUtilities.getMaxResultTime(getConnection());
    }

    @Override
    protected DateTime getMinResultTime() {
        return HibernateCriteriaQueryUtilities.getMinResultTime(getConnection());
    }

    @Override
    protected DateTime getMaxPhenomenonTime() {
        return HibernateCriteriaQueryUtilities.getMaxPhenomenonTime(getConnection());
    }

    @Override
    protected DateTime getMinPhenomenonTime() {
        return HibernateCriteriaQueryUtilities.getMinPhenomenonTime(getConnection());
    }

    @Override
    protected DateTime getMaxDateForOffering(final String offering) {
        return HibernateCriteriaQueryUtilities.getMaxDate4Offering(offering, getConnection());
    }

    @Override
    protected DateTime getMinResultTimeForOffering(final String offering) {
        return HibernateCriteriaQueryUtilities.getMinResultTime4Offering(offering, getConnection());
    }

    @Override
    protected DateTime getMaxResultTimeForOffering(final String offering) {
        return HibernateCriteriaQueryUtilities.getMaxResultTime4Offering(offering, getConnection());
    }

    @Override
    protected DateTime getMinDateForOffering(final String offering) {
        return HibernateCriteriaQueryUtilities.getMinDate4Offering(offering, getConnection());
    }

    @Override
    protected Session getConnection() {
        return session;
    }

    @Override
    protected void prepare() throws OwsExceptionReport {
        this.session = this.sessionHolder.getSession();
    }

    @Override
    protected void cleanup() {
        this.sessionHolder.returnSession(session);
    }
}
