/**
 * Copyright (C) 2013
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 */
package org.n52.sos.ds.hibernate;

import java.sql.Timestamp;
import java.util.List;

import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.criterion.Projections;
import org.hibernate.criterion.Restrictions;
import org.hibernate.transform.ResultTransformer;
import org.joda.time.DateTime;
import org.n52.sos.ds.hibernate.entities.FeatureOfInterest;
import org.n52.sos.ds.hibernate.entities.ObservableProperty;
import org.n52.sos.ds.hibernate.entities.Observation;
import org.n52.sos.ds.hibernate.entities.Procedure;
import org.n52.sos.exception.ows.NoApplicableCodeException;
import org.n52.sos.gda.AbstractGetDataAvailabilityDAO;
import org.n52.sos.gda.GetDataAvailabilityRequest;
import org.n52.sos.gda.GetDataAvailabilityResponse;
import org.n52.sos.gda.GetDataAvailabilityResponse.DataAvailability;
import org.n52.sos.ogc.gml.time.TimePeriod;
import org.n52.sos.ogc.ows.OwsExceptionReport;
import org.n52.sos.ogc.sos.SosConstants;

/**
 * {@code IGetDataAvailabilityDao} to handle {@link GetDataAvailabilityRequest}s.
 * 
 * @author Christian Autermann
 */
public class GetDataAvailabilityDAO extends AbstractGetDataAvailabilityDAO {

    private HibernateSessionHolder sessionHolder = new HibernateSessionHolder();
    
    public GetDataAvailabilityDAO() {
        super(SosConstants.SOS);
    }

    @Override
    public GetDataAvailabilityResponse getDataAvailability(GetDataAvailabilityRequest req)
            throws OwsExceptionReport {
        Session session = sessionHolder.getSession();
        try {
            Criteria c = session.createCriteria(Observation.class)
                    .add(Restrictions.eq(Observation.DELETED, false));

            if (req.isSetFeaturesOfInterest()) {
                c.createCriteria(Observation.FEATURE_OF_INTEREST)
                        .add(Restrictions.in(FeatureOfInterest.IDENTIFIER,
                                             req.getFeaturesOfInterest()));
            }
            if (req.isSetProcedures()) {
                c.createCriteria(Observation.PROCEDURE)
                        .add(Restrictions.in(Procedure.IDENTIFIER,
                                             req.getProcedures()));
                        
            }
            if (req.isSetObservedProperties()) {
                c.createCriteria(Observation.OBSERVABLE_PROPERTY)
                        .add(Restrictions.in(ObservableProperty.IDENTIFIER,
                                             req.getObservedProperties()));
            }

            c.setProjection(Projections.projectionList()
                    .add(Projections.groupProperty(Observation.PROCEDURE))
                    .add(Projections.groupProperty(Observation.OBSERVABLE_PROPERTY))
                    .add(Projections.groupProperty(Observation.FEATURE_OF_INTEREST))
                    .add(Projections.min(Observation.PHENOMENON_TIME_START))
                    .add(Projections.max(Observation.PHENOMENON_TIME_END)));

            c.setResultTransformer(new DataAvailabilityTransformer());

            List<?> dataAvailabilityValues = c.list();
            GetDataAvailabilityResponse response = new GetDataAvailabilityResponse();
            for (Object o : dataAvailabilityValues) {
                response.addDataAvailability((DataAvailability) o);
            }
            return response;
        } catch (HibernateException he) {
            throw new NoApplicableCodeException().causedBy(he)
                    .withMessage("Error while querying data for GetDataAvailability!");
        } finally {
            sessionHolder.returnSession(session);
        }
    }

    /**
     * Class to transform ResultSets to DataAvailabilities.
     */
    private static class DataAvailabilityTransformer implements ResultTransformer {
        private static final long serialVersionUID = -373512929481519459L;

        @Override
        public DataAvailability transformTuple(Object[] tuple, String[] aliases) {
            return new DataAvailability(((Procedure) tuple[0]).getIdentifier(),
                                        ((ObservableProperty) tuple[1]).getIdentifier(),
                                        ((FeatureOfInterest) tuple[2]).getIdentifier(),
                                        new TimePeriod(new DateTime(((Timestamp) tuple[3]).getTime()),
                                                       new DateTime(((Timestamp) tuple[4]).getTime())));
        }

        @Override
        @SuppressWarnings("rawtypes")
        public List transformList(List collection) {
            return collection;
        }
    }
}
