/**
 * Copyright (C) 2013
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 */
package org.n52.sos.gda;

import java.io.ByteArrayOutputStream;
import java.io.UnsupportedEncodingException;
import java.util.Collections;
import java.util.Map;
import java.util.Set;

import javax.xml.stream.XMLStreamException;

import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;
import org.n52.sos.encode.Encoder;
import org.n52.sos.encode.EncoderKey;
import org.n52.sos.exception.ows.NoApplicableCodeException;
import org.n52.sos.exception.ows.concrete.DateTimeFormatException;
import org.n52.sos.ogc.ows.OwsExceptionReport;
import org.n52.sos.ogc.sos.Sos2Constants;
import org.n52.sos.ogc.sos.SosConstants;
import org.n52.sos.ogc.sos.SosConstants.HelperValues;
import org.n52.sos.service.ServiceConstants.SupportedTypeKey;
import org.n52.sos.util.CodingHelper;
import org.n52.sos.util.StringHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Encoder} to handle {@link GetDataAvailabilityResponse}s.
 * 
 * @author Christian Autermann
 */
public class GetDataAvailabilityEncoder implements Encoder<XmlObject, GetDataAvailabilityResponse> {

    private static final Logger LOG = LoggerFactory.getLogger(GetDataAvailabilityEncoder.class);
    
    private static final Set<EncoderKey> ENCODER_KEYS = 
            CodingHelper.encoderKeysForElements(Sos2Constants.NS_SOS_20, GetDataAvailabilityResponse.class);
	
	/**
	 * Constructs a new {@code GetDataAvailabilityEncoder}.
	 */
	public GetDataAvailabilityEncoder() {
        LOG.debug("Encoder for the following keys initialized successfully: {}!", StringHelper.join(", ", ENCODER_KEYS));
	}
    
    @Override
    public Set<EncoderKey> getEncoderKeyType() {
        return Collections.unmodifiableSet(ENCODER_KEYS);
    }
    
    @Override
	public Map<SupportedTypeKey, Set<String>> getSupportedTypes() {
		return Collections.emptyMap();
	}

	@Override
	public void addNamespacePrefixToMap(Map<String, String> nameSpacePrefixMap) {
		nameSpacePrefixMap.put(Sos2Constants.NS_SOS_20, SosConstants.NS_SOS_PREFIX);
	}

	@Override
	public String getContentType() {
		return SosConstants.CONTENT_TYPE_XML;
	}

	@Override
	public Set<String> getConformanceClasses() {
		return Collections.emptySet();
	}

	@Override
    public XmlObject encode(GetDataAvailabilityResponse objectToEncode) throws OwsExceptionReport {
		return encode(objectToEncode, Collections.singletonMap(HelperValues.VERSION, Sos2Constants.SERVICEVERSION));
	}

	@Override
    public XmlObject encode(GetDataAvailabilityResponse gda, Map<HelperValues, String> additionalValues)
            throws OwsExceptionReport {
        try {
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            new GetDataAvailabilityStreamWriter(additionalValues.get(HelperValues.VERSION),
                                                gda.getDataAvailabilities()).write(out);
            return XmlObject.Factory.parse(out.toString("UTF8"));
        } catch (XMLStreamException ex) {
            throw new NoApplicableCodeException().causedBy(ex).withMessage("Error encoding response");
        } catch (DateTimeFormatException ex) {
            throw new NoApplicableCodeException().causedBy(ex).withMessage("Error encoding response");
        } catch (XmlException ex) {
            throw new NoApplicableCodeException().causedBy(ex).withMessage("Error encoding response");
        } catch (UnsupportedEncodingException ex) {
            throw new NoApplicableCodeException().causedBy(ex).withMessage("Error encoding response");
        }
    }

//     public static void main(String[] args) throws OwsExceptionReport {
//        TimePeriod tp = new TimePeriod(new DateTime(), new DateTime().plus(2000));
//        TimePeriod tp2 = new TimePeriod(new DateTime().minus(200000), new DateTime().plus(2000));
//        GetDataAvailabilityResponse r = new GetDataAvailabilityResponse(
//                new GetDataAvailabilityResponse.DataAvailability("foi_1", "obs_1", "proc_1", tp),
//                new GetDataAvailabilityResponse.DataAvailability("foi_2", "obs_2", "proc_2", tp),
//                new GetDataAvailabilityResponse.DataAvailability("foi_1", "obs_2", "proc_1", tp),
//                new GetDataAvailabilityResponse.DataAvailability("foi_2", "obs_1", "proc_2", tp2));
//         System.out.println(new GetDataAvailabilityEncoder().encode(r).xmlText(new XmlOptions()
//                 .setSaveAggressiveNamespaces().setSavePrettyPrint()));
//    }
}
